#include <linux/videodev2.h>
#include <media/msm_camera.h>
#include <media/msm_camera_sensor.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <sys/mman.h>
#include <linux/ioctl.h>

#include "camera.h"
#include "comdef.h"
#include "cam_mmap.h"

#include <gst/gst.h>

GST_DEBUG_CATEGORY_EXTERN(v4l2src_debug);
#define GST_CAT_DEFAULT v4l2src_debug

#define MSM_V4L2_START_SNAPSHOT _IOWR('V', BASE_VIDIOC_PRIVATE+1, \
      struct v4l2_buffer)

#define MSM_V4L2_GET_PICTURE    _IOWR('V', BASE_VIDIOC_PRIVATE+2, \
      struct v4l2_buffer)

#define MSM_V4L2_STOP_SNAPSHOT _IOWR('V', BASE_VIDIOC_PRIVATE+3, \
      struct v4l2_buffer)

static int ov265x_fd = -1;
/*===========================================================================
FUNCTION      v4l2_set_dimension 

DESCRIPTION
===========================================================================*/
boolean v4l2_set_dimension (
	int videofd, 
	void *pDim)
{
	struct v4l2_format vfmt;
	/* struct msm_v4l2_device_t * */
	
	boolean rc = TRUE;
	
	memset(&vfmt, 0, sizeof(struct v4l2_format) );
	cam_ctrl_dimension_t *pDimension = (cam_ctrl_dimension_t *)pDim;
	
	vfmt.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	vfmt.fmt.pix.width       = pDimension->display_width;
	vfmt.fmt.pix.height      = pDimension->display_height;
	
	vfmt.fmt.pix.priv  = (__u32)pDimension;
	
	
	vfmt.fmt.pix.pixelformat = V4L2_PIX_FMT_NV12;
	
	if (ioctl(videofd, VIDIOC_S_FMT, &vfmt) < 0) {
		GST_DEBUG("v4l2_set_dimension: VIDIOC_S_FMT failed");
		rc = FALSE;
	}
	
	pDimension->orig_picture_dx = pDimension->picture_width;
	pDimension->orig_picture_dy = pDimension->picture_height;
	
	pDimension->thumbnail_height = pDimension->ui_thumbnail_height;
	pDimension->thumbnail_width  = pDimension->ui_thumbnail_width;
		
	return rc;
}


/*===========================================================================
FUNCTION      native_set_default_params 

DESCRIPTION   
===========================================================================*/
boolean v4l2_set_default_params(int videofd)
{
	struct v4l2_control ctrlCmd;
	boolean rc = TRUE;
	
	memset(&ctrlCmd, 0, sizeof(struct v4l2_control) );
	
	ctrlCmd.id    = V4L2_CID_BRIGHTNESS;
	ctrlCmd.value = 0;
	if (ioctl(videofd, VIDIOC_S_CTRL, &ctrlCmd) < 0) {
		GST_DEBUG("v4l2_set_default_params: VIDIOC_S_CTRL: V4L2_CID_BRIGHTNESS failed...");
		rc = FALSE;
	}
	
	ctrlCmd.id    = V4L2_CID_CONTRAST;
	ctrlCmd.value = 0;
	if (ioctl(videofd, VIDIOC_S_CTRL, &ctrlCmd) < 0) {
		GST_DEBUG("v4l2_set_default_params: VIDIOC_S_CTRL: V4L2_CID_CONTRAST failed...");
		rc = FALSE;
	}
	
	ctrlCmd.id    = V4L2_CID_SATURATION;
	ctrlCmd.value = 0;
	if (ioctl(videofd, VIDIOC_S_CTRL, &ctrlCmd) < 0) {
		GST_DEBUG("v4l2_set_default_params: VIDIOC_S_CTRL: V4L2_CID_SATURATION failed...");
		rc = FALSE;
	}
	
	ctrlCmd.id    = V4L2_CID_HUE;
	ctrlCmd.value = 0;
	if (ioctl(videofd, VIDIOC_S_CTRL, &ctrlCmd) < 0) {
		GST_DEBUG("v4l2_set_default_params: VIDIOC_S_CTRL: V4L2_CID_HUE failed...");
		rc = FALSE;
	}
	
	return rc;
}

/*===========================================================================
FUNCTION      v4l2_register_preview_bufs

DESCRIPTION
===========================================================================*/
boolean v4l2_register_preview_bufs (
	int videofd, 
	int index,
	void *pDim,
	unsigned int addr,
	int fd,
	boolean active)
{
	struct v4l2_buffer  vb;
	uint32 prev_buf_size;

	GST_DEBUG("v4l2_register_preview_bufs: fd=%d buffer=%p", fd, addr);
	
	cam_ctrl_dimension_t *dimension = (cam_ctrl_dimension_t *)pDim;
	
	prev_buf_size = dimension->display_width *
					dimension->display_height;
	
	memset(&vb, 0, sizeof(struct v4l2_buffer));
	vb.index     = index;
	vb.type      = V4L2_BUF_TYPE_PRIVATE;
	vb.bytesused = prev_buf_size;
	vb.memory    = V4L2_MEMORY_MMAP;
	vb.m.userptr = addr;
	vb.flags     = 0x00010000;
	vb.reserved  = (uint32)fd;

	if (active)
		vb.flags |= 0x01;
	
	if (ioctl(videofd, VIDIOC_QBUF, &vb) == -1) {
		GST_DEBUG("v4l2_register_preview_bufs: VIDIOC_QBUF failed. %d", __LINE__);
		return FALSE;
	}
	return TRUE;
}

/*===========================================================================
FUNCTION      v4l2_unregister_preview_bufs

DESCRIPTION
===========================================================================*/
boolean v4l2_unregister_preview_bufs(
	int videofd, 
	void *pDim,
	int pmempreviewfd, 
	byte *prev_buf)
{
	struct v4l2_buffer  vb;
	
	memset(&vb, 0, sizeof(struct v4l2_buffer));
	vb.type      = V4L2_BUF_TYPE_PRIVATE;
	vb.m.userptr = prev_buf;
	vb.flags     = 0x00010000;
	vb.reserved  = (uint32)pmempreviewfd;
	
	if (ioctl(videofd, VIDIOC_DQBUF, &vb) == -1) {
		GST_DEBUG("v4l2_unregister_preview_bufs: VIDIOC_DQBUF failed. %d", __LINE__);
		return FALSE;
	}
	
	return TRUE;
}


/*===========================================================================
FUNCTION      v4l2_start_preview


DESCRIPTION
===========================================================================*/
boolean v4l2_start_preview(int videofd)
{
	uint16 ioctlRetVal = TRUE;
	enum v4l2_buf_type t = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	
	if (ioctl(videofd, VIDIOC_STREAMON, &t) == -1) {
		GST_DEBUG("v4l2_start_preview: ioctl failed.VIDIOC_STREAMON ");
		ioctlRetVal = -1;
		return FALSE;
	}

	return TRUE;
}


/*===========================================================================
FUNCTION      v4l2_stop_preview

DESCRIPTION
===========================================================================*/
boolean v4l2_stop_preview(int videofd)
{
	int ioctlRetVal = TRUE;
	enum v4l2_buf_type t = V4L2_BUF_TYPE_VIDEO_CAPTURE;
		
	if (ioctl(videofd, VIDIOC_STREAMOFF, &t) == -1) {
		GST_DEBUG("v4l2_stop_preview: ioctl failed.VIDIOC_STREAMON ");
		return FALSE;
	}
	
	return ioctlRetVal;
}


/*===========================================================================
FUNCTION      v4l2_start_snapshot

DESCRIPTION
===========================================================================*/
boolean v4l2_start_snapshot (int videofd)
{
	int ioctlRetVal;
	boolean rc = TRUE;
	
	enum v4l2_buf_type t = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	
	if ((ioctlRetVal = ioctl(videofd, MSM_V4L2_START_SNAPSHOT, &t)) < 0) { 
		GST_DEBUG("v4l2_start_snapshot: ioctl failed. ioctl return value is %d ", ioctlRetVal);
		rc = FALSE;
	}
	
	return rc;
}


/*===========================================================================
FUNCTION      v4l2_allocate_snapshot_bufs

DESCRIPTION
===========================================================================*/
boolean v4l2_register_snapshot_bufs (
	int videofd, 
	void *pDim, 
	int pmemThumbnailfd,
	int pmemSnapshotfd, 
	byte *thumbnail_buf, 
	byte *main_img_buf) 
{
	struct v4l2_buffer  thumbnailVb, snapshotVb;
	uint32 main_img_buf_size, thumbnail_buf_size;	
	cam_ctrl_dimension_t *dimension = (cam_ctrl_dimension_t *)pDim;
	
	main_img_buf_size  = dimension->picture_width *
						dimension->picture_height;
	
	thumbnail_buf_size = dimension->thumbnail_height * 
						dimension->thumbnail_width;
	
	GST_DEBUG("v4l2_register_snapshot_bufs: %d %d %d %d",
			dimension->picture_width, dimension->picture_height,
			dimension->thumbnail_height, dimension->thumbnail_width);
	
	memset(&thumbnailVb, 0, sizeof(struct v4l2_buffer));
	thumbnailVb.index     = 1;
	thumbnailVb.type      = V4L2_BUF_TYPE_PRIVATE;
	thumbnailVb.bytesused = thumbnail_buf_size;
	thumbnailVb.memory    = V4L2_MEMORY_MMAP;
	thumbnailVb.m.userptr = (unsigned long)thumbnail_buf;
	thumbnailVb.flags     = 0x00000101;
	thumbnailVb.reserved  = pmemThumbnailfd;
	
	if (ioctl(videofd, VIDIOC_QBUF, &thumbnailVb) == -1) {
		GST_DEBUG("v4l2_register_snapshot_bufs: VIDIOC_QBUF failed. %d", __LINE__);
		return FALSE;
	}
	
	memset(&snapshotVb, 0, sizeof(struct v4l2_buffer));
	snapshotVb.index     = 1;
	snapshotVb.type      = V4L2_BUF_TYPE_PRIVATE;
	snapshotVb.bytesused = main_img_buf_size;
	snapshotVb.memory    = V4L2_MEMORY_MMAP;
	snapshotVb.m.userptr = (unsigned long)main_img_buf;
	snapshotVb.flags     = 0x00000201;
	snapshotVb.reserved  = pmemSnapshotfd;
	
	if (ioctl(videofd, VIDIOC_QBUF, &snapshotVb) == -1) {
		GST_DEBUG("v4l2_register_snapshot_bufs: VIDIOC_QBUF failed %d", __LINE__);
		return FALSE;
	}
	return TRUE;
}


/*===========================================================================
FUNCTION      v4l2_unregister_snapshot_bufs

DESCRIPTION
===========================================================================*/
boolean v4l2_unregister_snapshot_bufs(
	int videofd, 
	void *pDim,
	int pmemThumbnailfd, 
	int pmemSnapshotfd, 
	byte *thumbnail_buf, 
	byte *main_img_buf)
{
	struct v4l2_buffer  vb;
	
	memset(&vb, 0, sizeof(struct v4l2_buffer));
	
	// Main image
	vb.type      = V4L2_BUF_TYPE_PRIVATE;
	vb.m.userptr = main_img_buf;
	vb.flags     = 0x00000201;
	vb.reserved  = (uint32)pmemSnapshotfd;
	
	if (ioctl(videofd, VIDIOC_DQBUF, &vb) == -1) {
		GST_DEBUG("v4l2_unregister_snapshot_bufs: VIDIOC_DQBUF failed. %d", __LINE__);
		return FALSE;
	}
	
	// Thumbnail
	vb.type      = V4L2_BUF_TYPE_PRIVATE;
	vb.m.userptr = thumbnail_buf;
	vb.flags     = 0x00000101;
	vb.reserved  = (uint32)pmemThumbnailfd;
	
	if (ioctl(videofd, VIDIOC_DQBUF, &vb) == -1) {
		GST_DEBUG("v4l2_unregister_snapshot_bufs: VIDIOC_DQBUF failed. %d", __LINE__);
		return FALSE;
	}
	
	return TRUE;
}




/*===========================================================================
FUNCTION      v4l2_get_picture

DESCRIPTION
===========================================================================*/
boolean v4l2_get_picture (int videofd)
{
	boolean rc = TRUE;
	struct msm_ctrl_cmd_t ctrlCmd;
	
	ctrlCmd.timeout_ms = 50000;
	ctrlCmd.length     = 0;
	ctrlCmd.value      = NULL;
	
	if(ioctl(videofd, MSM_V4L2_GET_PICTURE, &ctrlCmd) < 0) {
		GST_DEBUG("v4l2_get_picture:%d ioctl MSM_V4L2_GET_PICTURE FAILED...", __LINE__);
		rc = FALSE;
	}
	return rc;
}

/*===========================================================================
FUNCTION      v4l2_stop_snapshot 

DESCRIPTION
===========================================================================*/
boolean v4l2_stop_snapshot (int videofd)
{
	int ioctlRetVal;
	boolean rc = TRUE;
	
	enum v4l2_buf_type t = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	
	if((ioctlRetVal = ioctl(videofd, MSM_V4L2_STOP_SNAPSHOT, &t)) < 0) { 
		GST_DEBUG("v4l2_start_snapshot: ioctl failed. ioctl return value is %d ", ioctlRetVal);
		rc = FALSE;
	}

	return rc;
}

/*===========================================================================
 * FUNCTION    - v4l2_get_frame -  
 *
 * DESCRIPTION:
 *==========================================================================*/

int v4l2_get_frame(int videofd, unsigned int* addr, unsigned int* cbcr_off, int* fd, struct timeval* ts)
{
	struct v4l2_buffer vb;
	fd_set fds;
	int    rc = 0;
	struct timeval tv;
	
	GST_DEBUG("v4l2_get_frame");
	GST_DEBUG("v4l2_get_frame: set  time out to 80 milli seconds \n");
		
    	tv.tv_sec = 0;
	tv.tv_usec = 80000;

	FD_ZERO(&fds);
	FD_SET(videofd, &fds);

	rc = select(videofd + 1, &fds, NULL, NULL, &tv);
	
	if (0 > rc) {
		GST_ERROR("v4l2_get_frame: select failed...");
		return rc;
	} else if(0 == rc) {
		GST_DEBUG ("v4l2_get_frame:  timeout waiting for select! %d ",ETIMEDOUT );
		return -ETIMEDOUT;
	}

	memset(&vb, 0, sizeof(struct v4l2_buffer));
	vb.type   = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	vb.memory = V4L2_MEMORY_MMAP;

	// Get new buffer from camera
	rc = ioctl(videofd, VIDIOC_DQBUF, &vb);
	
	if (!rc) {
		*addr			= vb.m.userptr;
		*cbcr_off		= vb.bytesused;
		*fd				= vb.reserved;
		*ts				= vb.timestamp;

		GST_DEBUG("v4l2_get_frame: fd=%d addr=%p", *fd, *addr);
	}
	else {
		GST_ERROR("v4l2_get_frame: VIDIOC_DQBUF failed! %d", __LINE__);
	}

	return rc;
}

/*===========================================================================
 * FUNCTION    - v4l2_put_frame -  
 *
 * DESCRIPTION:
 *==========================================================================*/

boolean v4l2_put_frame(int videofd, unsigned int addr, unsigned int cbcr_off, int fd)
{
	struct v4l2_buffer vb;
	int    rc = 0;

	GST_DEBUG("v4l2_put_frame: fd=%d addr=%p", fd, addr);

	memset(&vb, 0, sizeof(struct v4l2_buffer));
	vb.type   = V4L2_BUF_TYPE_VIDEO_CAPTURE;
	vb.memory = V4L2_MEMORY_MMAP;
	vb.m.userptr  = addr;
	vb.reserved   = fd;
	vb.bytesused  = cbcr_off;
			
	// Queue the buffer back to the camera
	rc = ioctl(videofd, VIDIOC_QBUF, &vb);

	if (rc < 0) {
		GST_DEBUG("v4l2_put_frame: VIDIOC_QBUF failed! %d", __LINE__);
	}

	return rc;
}

/*===========================================================================
FUNCTION      v4l2_interface_destroy

DESCRIPTION
===========================================================================*/
boolean v4l2_interface_destroy(int videofd)
{
	GST_DEBUG("v4l2_interface_destroy");

	close(videofd); // handle to /dev/video20
	close(ov265x_fd);

	return TRUE;
}

/*===========================================================================
FUNCTION      v4l2_interface_init

DESCRIPTION
===========================================================================*/
boolean v4l2_interface_init(int *videofd)
{
	*videofd = open( "/dev/video20", O_RDWR | O_NONBLOCK, 0 );
	if (*videofd == -1) {
		GST_DEBUG("v4l2_interface_init: failed to open /dev/video");
		return FALSE;
	}
	
	ov265x_fd = open ("/dev/ov265x", O_RDWR | O_NONBLOCK, 0);
  	if (ov265x_fd == -1) {
    		GST_DEBUG("v4l2_interface_init: failed to open /dev/ov265x");
		close (*videofd);
    		return FALSE;
  	}

	return TRUE;
}

/*===========================================================================
FUNCTION      v4l2_readregister

DESCRIPTION	Function that returns the sensor register value, given the sensor 			register address
===========================================================================*/


int v4l2_readregister (int register_address){

  int ret;
  int rc = -1;
  struct sensor_cfg_data_t cfg_data;

  memset(&cfg_data, 0, sizeof(struct sensor_cfg_data_t) );

  cfg_data.cfg.reg.register_address = register_address;
  if((ret = ioctl(ov265x_fd, MSM_CAMERA_READ_SENSOR_REGISTER, &cfg_data)) < 0) { 
       	GST_ERROR( "v4l2_readregister: ioctl failed. ioctl return value is %d \n", ret);
       	rc = -1;
	
  }else{

  	rc = cfg_data.cfg.reg.register_value;
	
   }
	

  return rc;
  

}

/*===========================================================================
FUNCTION      v4l2_cameraflashcontrol

DESCRIPTION		Function that turns on the flash, given the current
===========================================================================*/
boolean v4l2_cameraflashcontrol (int current)
{
  int ioctlRetVal;
  boolean rc = TRUE;
  struct sensor_cfg_data_t cfg_data;

  memset(&cfg_data, 0, sizeof(struct sensor_cfg_data_t) );

  cfg_data.cfg.flash.flash_led_mode_type = 0; // type
  cfg_data.cfg.flash.flash_led_polarity_type = 0; // polarity
  cfg_data.cfg.flash.flash_led_current_ma = current;

  if((ioctlRetVal = ioctl(ov265x_fd, MSM_CAMERA_CONFIG_CURRENT, &cfg_data)) < 0) { 
    GST_ERROR("v4l2_cameraflashcontrol: ioctl failed. ioctl return value is %d \n", ioctlRetVal);
    rc = FALSE;
  }

  return rc;
}




/*===========================================================================
FUNCTION      v4l2_cameraflashcontrol_ex

DESCRIPTION		Function that turns on the flash, given the current, flashmode
===========================================================================*/
boolean v4l2_cameraflashcontrol_ex (int current, int flash_mode)
{
  int ioctlRetVal;
  boolean rc = TRUE;
  struct sensor_cfg_data_t cfg_data;

  memset(&cfg_data, 0, sizeof(struct sensor_cfg_data_t) );

  cfg_data.cfg.flash.flash_led_mode_type = 0; // type
  cfg_data.cfg.flash.flash_led_polarity_type = 0; // polarity
  cfg_data.cfg.flash.flash_led_current_ma = current;
  cfg_data.cfg.flash.cam_flash_mode = flash_mode;
  if((ioctlRetVal = ioctl(ov265x_fd, MSM_CAMERA_CONFIG_CURRENT_EX, &cfg_data)) < 0) { 
    GST_ERROR("v4l2_cameraflashcontrol: ioctl failed. ioctl return value is %d \n", ioctlRetVal);
    rc = FALSE;
  }

  return rc;
}


